"""
Stored data for use during runtime or testing.

Because dbCAN is an evolving database, we'll have to maintain several models
for each database release. The latest version will always we the default one.
"""

from __future__ import unicode_literals

from pkg_resources import resource_filename

from collections import defaultdict
import json
from enum import Enum

import numpy as np

from catas.matrix import Matrix


class MyEnum(Enum):
    """ Base class for enums. """

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s):
        try:
            return cls[s]
        except KeyError:
            raise ValueError

    @classmethod
    def from_other(cls, f):
        if isinstance(f, cls):
            return f
        elif isinstance(f, str):
            return cls.from_string(f)
        elif isinstance(f, int):
            return cls(f)
        else:
            raise ValueError("Expected an enum, string or integer.")


class Version(MyEnum):
    """ Supported DBCAN database versions. """

    v4 = 1
    v5 = 2
    v6 = 3
    v7 = 4

    @classmethod
    def latest(cls):
        return cls.v6


class Nomenclature(MyEnum):
    """ All CATAStrophy nomenclatures. """

    nomenclature1 = 1
    nomenclature2 = 2
    nomenclature3 = 3

    @classmethod
    def default(cls):
        return cls.nomenclature3


def models(version=Version.latest()):
    """ Loads PCA parameters. """

    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-model.npz"),
        (Version.v5, "v5-20190311-model.npz"),
        (Version.v6, "v6-20190311-model.npz"),
        (Version.v7, "v7-20190311-model.npz"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    return np.load(version_files[version])


def principle_components(version=Version.latest()):
    """ Loads files containing PCs for training data. """

    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-principle_components.npz"),
        (Version.v5, "v5-20190311-principle_components.npz"),
        (Version.v6, "v6-20190311-principle_components.npz"),
        (Version.v7, "v7-20190311-principle_components.npz"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    mat_raw = np.load(version_files[version])
    mat = Matrix.read(version_files[version])

    mat.nomenclature1 = mat_raw["nomenclature1"]
    mat.nomenclature2 = mat_raw["nomenclature2"]
    mat.nomenclature3 = mat_raw["nomenclature3"]
    return mat


def cazy_list(version=Version.latest()):
    """ Loads CAZY list of class names the in order that model was trained. """

    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-cazy_list.json"),
        (Version.v5, "v5-20190311-cazy_list.json"),
        (Version.v6, "v6-20190311-cazy_list.json"),
        (Version.v7, "v7-20190311-cazy_list.json"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    with open(version_files[version], "r") as handle:
        d = json.load(handle)

    return d


def centroids(version=Version.latest(), nomenclature=Nomenclature.default()):
    """ Loads pandas dataframe containing centroids. """

    # Handle strings as input instead of just enums.
    version = Version.from_other(version)
    nomenclature = Nomenclature.from_other(nomenclature)

    files = defaultdict(dict)
    filelist = [
        (
            Version.v4,
            Nomenclature.nomenclature1,
            "v4-20190311-nomenclature1-centroids.npz"
        ),
        (
            Version.v4,
            Nomenclature.nomenclature2,
            "v4-20190311-nomenclature2-centroids.npz"
        ),
        (
            Version.v4,
            Nomenclature.nomenclature3,
            "v4-20190311-nomenclature3-centroids.npz"
        ),
        (
            Version.v5,
            Nomenclature.nomenclature1,
            "v5-20190311-nomenclature1-centroids.npz"
        ),
        (
            Version.v5,
            Nomenclature.nomenclature2,
            "v5-20190311-nomenclature2-centroids.npz"
        ),
        (
            Version.v5,
            Nomenclature.nomenclature3,
            "v5-20190311-nomenclature3-centroids.npz"
        ),
        (
            Version.v6,
            Nomenclature.nomenclature1,
            "v6-20190311-nomenclature1-centroids.npz"
        ),
        (
            Version.v6,
            Nomenclature.nomenclature2,
            "v6-20190311-nomenclature2-centroids.npz"
        ),
        (
            Version.v6,
            Nomenclature.nomenclature3,
            "v6-20190311-nomenclature3-centroids.npz"
        ),
        (
            Version.v7,
            Nomenclature.nomenclature1,
            "v7-20190311-nomenclature1-centroids.npz"
        ),
        (
            Version.v7,
            Nomenclature.nomenclature2,
            "v7-20190311-nomenclature2-centroids.npz"
        ),
        (
            Version.v7,
            Nomenclature.nomenclature3,
            "v7-20190311-nomenclature3-centroids.npz"
        ),
    ]

    for vname, nname, vfile in filelist:
        files[vname][nname] = resource_filename(__name__, vfile)

    return Matrix.read(files[version][nomenclature])


def hmm_lengths(version=Version.latest()):
    """ Loads dict object containing lengths of HMMs.

    The raw HMMER output doesn't contain the length of the HMM, which we
    need to compute coverage.
    """

    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-hmm_lengths.json"),
        (Version.v5, "v5-20190311-hmm_lengths.json"),
        (Version.v6, "v6-20190311-hmm_lengths.json"),
        (Version.v7, "v7-20190311-hmm_lengths.json"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    with open(version_files[version], "r") as handle:
        d = json.load(handle)

    return d


def trophic_classes(nomenclature=Nomenclature.default()):
    """ Loads list containing names and correct order to display classes. """

    nomenclature = Nomenclature.from_other(nomenclature)

    files = dict()
    filelist = [
        (Nomenclature.nomenclature1, "nomenclature1-trophic_classes.json"),
        (Nomenclature.nomenclature2, "nomenclature2-trophic_classes.json"),
        (Nomenclature.nomenclature3, "nomenclature3-trophic_classes.json"),
    ]

    for nname, nfile in filelist:
        files[nname] = resource_filename(__name__, nfile)

    with open(files[nomenclature], "r") as handle:
        d = json.load(handle)
    return d


def sample_fasta():
    # There's no actual version for this.
    return resource_filename(__name__, "test_data.fasta")


def test_dbcan(version=Version.latest()):
    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-test_dbcan.csv"),
        (Version.v5, "v5-20190311-test_dbcan.csv"),
        (Version.v6, "v6-20190311-test_dbcan.csv"),
        (Version.v7, "v7-20190311-test_dbcan.csv"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    return version_files[version]


def test_hmmer(version=Version.latest()):
    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-test_hmmer.txt"),
        (Version.v5, "v5-20190311-test_hmmer.txt"),
        (Version.v6, "v6-20190311-test_hmmer.txt"),
        (Version.v7, "v7-20190311-test_hmmer.txt"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    return version_files[version]


def test_domtblout(version=Version.latest()):
    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-test_hmmer.csv"),
        (Version.v5, "v5-20190311-test_hmmer.csv"),
        (Version.v6, "v6-20190311-test_hmmer.csv"),
        (Version.v7, "v7-20190311-test_hmmer.csv"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    return version_files[version]


def test_counts(version=Version.latest()):
    # Handle strings as input instead of just enums.
    version = Version.from_other(version)

    version_files = dict()
    filelist = [
        (Version.v4, "v4-20190311-test_counts.npz"),
        (Version.v5, "v5-20190311-test_counts.npz"),
        (Version.v6, "v6-20190311-test_counts.npz"),
        (Version.v7, "v7-20190311-test_counts.npz"),
    ]

    for vname, vfile in filelist:
        version_files[vname] = resource_filename(__name__, vfile)

    return Matrix.read(version_files[version])



from io import StringIO
from enum import Enum
from typing import NamedTuple

from Bio import SearchIO

from catas.data import Version
from catas.data import hmm_lengths


class FileType(Enum):
    """ All possible input filetypes that we support. """

    dbcan = 1
    hmmer_text = 2
    hmmer_domtab = 3

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s):
        try:
            return cls[s]
        except KeyError:
            raise ValueError


class ParseError(Exception):
    """ Some aspect of parsing failed. """

    def __init__(self, filename, line, message):
        self.filename = filename
        self.line = line
        self.message = message


class LineParseError(Exception):
    def __init__(self, message):
        self.message


def parse(handle, format, version=None):
    """ Wrapper that directs parsing bases on enum. """

    version = Version.from_other(version)

    if isinstance(format, str):
        format = FileType[format]

    if format == FileType.dbcan:
        return DBCAN.from_file(handle)
    elif format == FileType.hmmer_domtab:
        parsed = HMMER.from_file(
            handle,
            format="hmmscan3-domtab",
            version=version,
        )
        return parsed
    elif format == FileType.hmmer_text:
        return HMMER.from_file(
            handle,
            format="hmmer3-text",
            version=version,
        )
    else:
        raise ValueError((
            "Parse is only defined for 'dbcan', 'hmmscan3-domtab', "
            "'hmmer3-text'"
        ))


def decode_object(handle, num_lines=None):
    """ Converts a binary file-like object into a String one.

    Files transferred over http arrive as binary objects.
    If I want to check the first few lines I need to decode it.
    """

    # If we don't specify a number of lines, just use infinity
    if num_lines is None:
        num_lines = float("inf")
    lines = [l.decode() for i, l
             in enumerate(handle)
             if i < num_lines]

    # Create an in memory file like object
    out = StringIO()
    # Write the lines to the file like object.
    out.writelines(lines)

    # Rewind the files to the start so that we can read them
    out.seek(0)
    handle.seek(0)
    return out


def split_hmm(string):
    return str(string).rstrip(".hmm")


class DBCAN(NamedTuple):
    """ Stores a single row of the DBCAN tab-delimited file. """

    hmm: str
    hmm_len: int
    seqid: str
    query_len: int
    evalue: float
    hmm_from: int
    hmm_to: int
    ali_from: int
    ali_to: int
    coverage: float

    @classmethod
    def from_string(cls, line):
        """ Parse a single tabular output row from dbCAN. """

        columns = [split_hmm, int, str, int, float, int, int, int, int, float]
        line = line.strip().split('\t')

        if len(columns) != len(line):
            msg = "Wrong number of columns. Found {}, expected {}."
            raise LineParseError(msg.format(len(line), len(columns)))

        output = []
        for i, type_fn in enumerate(columns):
            try:
                output.append(type_fn(line[i]))
            except IndexError:
                # This shouldn't happen because of previous check
                msg = "Couldn't parse file as dbcan format, missing column."
                raise LineParseError(msg)
            except ValueError:
                msg = "Illegal value '{}' received in column {}."
                raise LineParseError(msg.format(line[i], i + 1))
        return cls(*output)

    @classmethod
    def from_file(cls, handle):
        """ Parse a DBCAN file into generator of rows. """

        for line_number, line in enumerate(handle, 1):
            line = line.strip()
            # Skip comment lines
            if line.startswith("#"):
                continue
            # Skip empty lines
            elif line == "":
                continue

            try:
                parsed = cls.from_string(line)
            except LineParseError as e:
                raise ParseError(handle.name, line_number, e.msg)

            yield parsed
        return


class HMMER(NamedTuple):
    """ Stores a single HMMER match. """

    hmm: str
    hmm_len: int
    seqid: str
    evalue: float
    hmm_from: int
    hmm_to: int
    ali_from: int
    ali_to: int
    coverage: float

    @classmethod
    def from_file(cls, handle, format="hmmer3-text", version=Version.latest()):
        """ Processes HMMER output in a similar way to dbCAN hmmscan-parser.sh.

        There are two main HMMER output types, the plaintext output and the
        domain table type output (provided by `--domtblout`). The plaintext
        format is not particularly friendly, so we're using the Biopython
        parser.  This has the added benefit of handling both the domain table
        and plaintext format with minimal alteration.

        Keyword arguments:
        handle -- A python file handle of the HMMER output.
        format -- A string specifying an input format supported by the
            Biopython SearchIO API. Should be either 'hmmer3-text' or
            'hmmscan3-domtab'.

        Returns:
        generator -- A generator yielding named tuples corresponding to lines.
        """

        version = Version.from_other(version)

        # This is the threshold used by dbcan to determine if to matches are
        # the same.
        overlap_thres = 0.5

        # Needed to find the coverage of hmms, some formats are missing this.
        hmm_lens = hmm_lengths(version)

        # Parse the file using Biopython.
        matches = SearchIO.parse(handle, format=format)

        # Loop through matches
        try:
            for query in matches:
                # For each HMM type that matches the sequence

                query_matches = list()
                for hit in query.hits:
                    # For each partial alignment of the HMM to the sequence.
                    # Multiple HSPs can occur if there are genuinely two
                    # copies of the domain or if a large domain is split
                    # into two alignments.
                    for hsp in hit.hsps:
                        # Drop rows with alignment lengths of zero
                        if hsp.query_end - hsp.query_start <= 0:
                            continue

                        # Format and yield the line elements that we want.
                        row = cls._get_hsp(query, hit, hsp, hmm_lens)
                        query_matches.append(row)
                # Get the best match for overlapping matches.
                query_matches.sort(key=lambda x: (x.ali_from, x.ali_to))
                query_winners = cls._distinct_matches(
                    query_matches,
                    overlap_thres
                )
                # Yield the rows.
                for winner in query_winners:
                    # If this winner passes the coverage and evalue thresholds
                    # yield it.
                    if cls._filter_significant(winner):
                        yield winner

        except AssertionError:
            msg = (
                "Couldn't parse file as {} format.\n"
                "Double check that the input file is in the right format "
                "(i.e. hmmer_text vs hmmer_domtab).\n"
                "If you believe this is an error, please contact us."
            ).format(format)
            raise ParseError(handle.name, None, msg)
        return

    @classmethod
    def _get_hsp(cls, query, hit, hsp, hmm_lens):
        """ Construct a record from a biopython SearchIO tree. """

        hmm = split_hmm(hit.id)
        hmm_len = hmm_lens[hmm]

        hmm_from = hsp.hit_start
        hmm_to = hsp.hit_end

        coverage = (hmm_to - hmm_from) / hmm_len
        return cls(
            hmm=hmm,
            hmm_len=hmm_len,
            seqid=query.id,
            evalue=hsp.evalue,
            hmm_from=hmm_from,
            hmm_to=hmm_to,
            ali_from=hsp.query_start,
            ali_to=hsp.query_end,
            coverage=coverage,
        )

    @staticmethod
    def _distinct_matches(matches, overlap_thres):
        # Accumulate winners into a list.
        winners = list()

        # Keep track of the current best hits.
        winner = None
        winner_ali_length = 0

        # Loop through each match for each seqid, replacing "winner" as we go.
        for match in matches:
            # If this is the first match winner will be None, so we just set
            # it to the first value and continue.
            if winner is None:
                winner = match
                winner_ali_length = winner.ali_to - winner.ali_from
                # Continue moves to the next match in the list.
                continue

            # Get the distance between current match and winner
            dist = winner.ali_to - match.ali_from

            # Decide whether or not the matches are the same
            match_ali_length = match.ali_to - match.ali_from
            winner_nooverlap = (dist / winner_ali_length) > overlap_thres
            match_nooverlap = (dist / match_ali_length) > overlap_thres

            # If the distance is positive (meaning that row start is before
            # current winner end) and the distance is a large proportion of
            # the alignment length, consider them to be the same.
            if dist > 0 and (winner_nooverlap or match_nooverlap):
                # If the new rows evalue is less than the current winners
                # evalue the new row becomes the new winner!
                if match.evalue < winner.evalue:
                    winner = match
                    winner_ali_length = match_ali_length
            else:
                # We're onto a new alignment and the previous winner gets
                # to stay.
                winners.append(winner)
                winner = match
                winner_ali_length = match_ali_length

        # At the end we'll have a new winner to add.
        # But not if the list was empty.
        if winner is not None:
            winners.append(winner)
        return winners

    @staticmethod
    def _filter_significant(match):
        """ Select significant alignments based on evalues and coverage.

        dbCAN uses a threshold of 1e-5 for alignments over 80 AA
        and 1e-3 for others. A coverage threshold of 0.3 is used.
        """

        if match.coverage < 0.3:
            return False

        is_long = (match.ali_to - match.ali_from) > 80

        if is_long:
            return match.evalue < 1e-5
        else:
            return match.evalue < 1e-3



"""
"""

import pytest

from catas.parsers import DBCAN
from catas.count import cazy_counts
from catas.data import Version
from catas.data import test_dbcan
from catas.data import cazy_list


@pytest.mark.parametrize("version,hmm,exp_val", [
    (Version.v4, "CBM1", 2),
    (Version.v4, "GH10", 4),
    (Version.v4, "GH100", 0),
    (Version.v5, "CBM1", 2),
    (Version.v5, "GH3", 3),
    (Version.v5, "CE10", 0),
    (Version.v6, "CBM1", 2),
    (Version.v6, "GH11", 1),
    (Version.v6, "GH31", 0),
])
def test_cazy_counts(version, hmm, exp_val):
    required_cols = cazy_list(version)
    with open(test_dbcan(version=version), "r") as handle:
        parsed = DBCAN.from_file(handle)
        counts = cazy_counts(parsed, required_cols)

    column_index = required_cols.index(hmm)
    assert counts[column_index] == exp_val
    return



"""
A simpler alternative to a pandas dataframe for non-interactive work.
"""

import csv
import numpy as np


class Matrix(object):
    """ A simple wrapper around a numpy array that tracks column and row names.

    Order is row major.
    Using this as an alternative to pandas to get rid of dependency.
    """

    def __init__(self, rows, columns, arr):
        assert len(rows) == arr.shape[0]
        assert len(columns) == arr.shape[1]

        self.rows = rows
        self.columns = columns
        self.arr = arr
        return

    def __getitem__(self, key):
        if isinstance(key, str):
            key = self.columns.index(key)

        return self.arr[:, key]

    @classmethod
    def from_row(cls, row, columns, arr):
        """ Create a matrix from a single row """

        return cls([row], columns, arr.reshape(1, -1))

    @classmethod
    def concat(cls, dfs):
        """ Concatenate multiple matrices together by rows (ie. columns are
        matched) """

        if len(dfs) == 0:
            return cls([], [], np.zeros((0, 0)))

        columns = dfs[0].columns

        # Columns must all be same shape and names.
        for df in dfs:
            assert df.columns == columns

        # Concatenates and flattens all rows.
        rows = [r for df in dfs for r in df.rows]
        # NB concatenate requires all dfs to have correct dimensions.
        # IE they can't be a flat array.
        arr = np.concatenate([df.arr for df in dfs])
        return cls(rows, columns, arr)

    def as_df(self):
        import pandas as pd
        return pd.DataFrame(
            data=self.arr,
            index=self.rows,
            columns=self.columns
        )

    @classmethod
    def from_df(cls, df):
        """ Assumes heterogeneous df with rownames. """
        rows = df.index.tolist()
        columns = df.columns.tolist()
        return cls(rows, columns, df.values)

    def write(self, handle):
        """ Writes matrix out as a numpy file. """
        np.savez(
            handle,
            arr=self.arr,
            rows=self.rows,
            columns=self.columns
        )
        return

    @classmethod
    def read(cls, handle):
        """ Reads matrix from an numpy file. """
        f = np.load(handle, allow_pickle=False)
        return cls(f["rows"].tolist(), f["columns"].tolist(), f["arr"])

    def write_tsv(self, filename, rows_colname="label"):
        """ Write a tsv formatted table from the matrix.

        Note that if you provide a file-object handle to filename, it must
        have been opened with the `newline=''` option (required by csv).
        """
        try:
            if isinstance(filename, str):
                handle = open(filename, 'w', newline='')
            else:
                handle = filename

            tab_writer = csv.writer(handle, "excel-tab")

            # Write the header
            tab_writer.writerow([rows_colname] + self.columns)

            # Write the lines
            for rowname, row in zip(self.rows, self.arr):
                tab_writer.writerow([rowname] + list(row))

        finally:
            if isinstance(filename, str):
                handle.close()
        return

    def as_serializable(self):
        """ Output the matrix as a dict, which can be easily serialised into
        JSON etc.
        """

        output = list()
        for rowname, row in zip(self.rows, self.arr):
            drow = {str(k): float(v) for k, v in zip(self.columns, row)}
            drow["label"] = rowname
            output.append(drow)
        return output

    @classmethod
    def from_serialized(cls, ser):
        """ Parses a dictionary representation back into a matrix. """

        if len(ser) == 0:
            return cls([], [], np.zeros((0, 0)))

        rows = list()
        arr = list()
        columns = [k for k, v in ser if k != "label"]
        for row in ser:
            arr_row = list()
            rows.append(row["label"])
            for column in columns:
                arr_row.append(row[column])
            arr.append(arr_row)

        return cls(
            rows,
            columns,
            np.array(arr),
        )



"""
Functions to find the PCA-transformed values from cazyme counts.
Functions to find distances to centroids in PCA space.
"""

import numpy as np
from numpy.linalg import norm

from catas.matrix import Matrix
from catas.data import models
from catas.data import centroids
from catas.data import Version
from catas.data import Nomenclature


def predict(
    counts,
    version=Version.latest(),
    nomenclature=Nomenclature.default()
):
    """ Wrapper function that takes file handle and returns RCD predictions.

    Keyword arguments:
    counts -- A `Matrix` containing the cazyme counts (required).
    version -- The model to run the predictions against (Default is latest
        available in data).
    nomenclature -- The nomenclature to predict labels for.

    Returns:
    Matrix -- A matrix of RCDs, with samples in rows and nomenclature labels
        as columns.
    """

    version = Version.from_other(version)
    nomenclature = Nomenclature.from_other(nomenclature)

    # load the sklearn pipeline that runs scaling and PCA
    model = models(version)

    # Load the DataFrame containing PCA centroids for each class.
    ctds = centroids(version=version, nomenclature=nomenclature)

    # Transform the counts into PCA space values.
    trans = transform(counts, model=model)

    # Find the distances between the point in PCA space and class centroids.
    dists = distances(trans, centroids=ctds)

    # Calculate the relative centroid distance for each distance.
    rcds = rcd(dists)

    return rcds


def transform(counts, model=models()):
    """ Takes a series of CAZyme counts and gets PCA transformed values.

    Keyword arguments:
    counts -- A Matrix containing CAZyme counts.
    model -- A dictionary containing numpy arrays used to transform the counts.
        E.g. from `catas.data.models()`.

    Returns:
    Matrix -- A matrix object with the PCA transformed values.
    """

    values = counts.arr.astype(np.float)
    X = values - model["mean"]
    X_transformed = np.dot(X, model["components"].T)

    columns = ["pc{:0>2}".format(i + 1) for i in range(X_transformed.shape[1])]
    return Matrix(rows=counts.rows, columns=columns, arr=X_transformed)


def distances(points, centroids=centroids()):
    """ Given a point in PCA space, find the distance to each centroid.

    Keyword arguments:
    points -- A Matrix representing the location of sample(s) in PC space.
    centroids -- A matrix with rows representing classes and columns
        representing class centroids in principle component space (Default is
        latest centroids available in data).
        E.g. from `catas.data.centroids()`.

    Returns:
    Matrix -- A Matrix object giving the euclidean distance to each
        class centroid indexed by the class names.
    """

    # This computes the distance matrix, with samples in rows and centroids as
    # columns.
    diffs = np.apply_along_axis(lambda x: x - centroids.arr, 1, points.arr)
    normed = np.apply_along_axis(norm, 2, diffs)

    # Grab the class names from the centroids
    new_columns = centroids.rows

    # return a new series with the class names as index
    return Matrix(rows=points.rows, columns=new_columns, arr=normed)


def rcd(dists):
    """ Finds the relative centroid distance.

    Given an array of distances between two points,
    returns the RCD for each distance in the array.

    Keyword arguments:
    dists -- A Matrix object representing distances between points and
        centroids.

    Returns:
    Matrix -- A Matrix object with the RCD values for each class.
    """

    min_ = dists.arr.min(axis=1).reshape(-1, 1)
    max_ = dists.arr.max(axis=1).reshape(-1, 1)
    ratio = (dists.arr - min_) / (max_ - min_)
    rcd_ = 1 - ratio

    # Return the RCDs as a series object with original names/index
    return Matrix(rows=dists.rows, columns=dists.columns, arr=rcd_)



"""
"""

import pytest
from numpy.testing import assert_almost_equal

from catas.parsers import DBCAN
from catas.count import cazy_counts_multi
from catas.predict import predict
from catas.predict import transform
from catas.predict import distances
from catas.predict import rcd

from catas.data import test_dbcan
from catas.data import models
from catas.data import centroids
from catas.data import Version
from catas.data import Nomenclature
from catas.data import cazy_list


@pytest.mark.parametrize("version,clss,exp_val", [
    (Version.v5, "biotroph 3", 0.9230486190150349),
    (Version.v5, "saprotroph", 0.8919083654247382),
    (Version.v5, "wilt", 0.0),
    (Version.v5, "mesotroph - internal", 0.22047031254890714),
    (Version.v4, "biotroph 3", 0.9233213567655258),
    (Version.v4, "saprotroph", 0.8913578386735521),
    (Version.v4, "wilt", 0.0),
    (Version.v4, "mesotroph - internal", 0.2201692033555083),
])
def test_predict(version, clss, exp_val):
    """ Pretty much a repeat of test_rcds. """

    with open(test_dbcan(version=version), "r") as handle:
        parsed = DBCAN.from_file(handle)
        required_cols = cazy_list(version)
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    rcds = predict(counts, version, Nomenclature["nomenclature3"])

    assert rcds.rows[0] == "test"

    column_index = rcds.columns.index(clss)
    # Checks that results are accurate to 5 decimal places.
    assert_almost_equal(exp_val, rcds.arr[0, column_index], decimal=5)
    return


@pytest.mark.parametrize("version,pc,exp_val", [
    (Version.v4, "pc01", -69.00896835173042),
    (Version.v4, "pc02", -14.168483882671623),
    (Version.v5, "pc01", -68.9655038492295),
    (Version.v5, "pc02", -14.158262281499532),
    (Version.v6, "pc01", -64.99073313231153),
    (Version.v6, "pc02", -13.146559438278567),
])
def test_pca(version, pc, exp_val):

    with open(test_dbcan(version=version), "r") as handle:
        parsed = DBCAN.from_file(handle)
        required_cols = cazy_list(version)
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    print(counts)
    model = models(version)
    trans = transform(counts, model=model)

    # Checks that results are accurate to 5 decimal places.
    column_index = trans.columns.index(pc)
    assert_almost_equal(exp_val, trans.arr[0, column_index], decimal=5)
    return


@pytest.mark.parametrize("version,clss,exp_val", [
    (Version.v5, "biotroph 2", 51.57963882216139),
    (Version.v5, "mesotroph - external", 70.20617822839525),
    (Version.v5, "wilt", 141.11596086281014)
])
def test_distances(version, clss, exp_val):
    with open(test_dbcan(version=version), "r") as handle:
        parsed = DBCAN.from_file(handle)
        required_cols = cazy_list(version)
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    model = models(version)
    trans = transform(counts, model=model)

    cent = centroids(version)
    dists = distances(trans, centroids=cent)

    # Checks that results are accurate to 5 decimal places.
    column_index = dists.columns.index(clss)
    assert_almost_equal(exp_val, dists.arr[0, column_index], decimal=5)
    return


@pytest.mark.parametrize("version,clss,exp_val", [
    (Version.v5, "biotroph 3", 0.9230486190150349),
    (Version.v5, "saprotroph", 0.8919083654247382),
    (Version.v5, "wilt", 0.0),
    (Version.v5, "mesotroph - internal", 0.22047031254890714),
])
def test_rcd(version, clss, exp_val):
    with open(test_dbcan(version=version), "r") as handle:
        parsed = DBCAN.from_file(handle)
        required_cols = cazy_list(version)
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    model = models(version)
    trans = transform(counts, model=model)

    cent = centroids(version)
    dists = distances(trans, centroids=cent)
    rcds = rcd(dists)

    # Checks that results are accurate to 5 decimal places.
    column_index = rcds.columns.index(clss)
    assert_almost_equal(exp_val, rcds.arr[0, column_index], decimal=5)
    return



import sys
import argparse
import traceback

from catas.predict import predict
from catas.count import cazy_counts_multi
from catas import parsers
from catas.parsers import FileType
from catas.parsers import ParseError
from catas.data import Version
from catas.data import Nomenclature
from catas.data import cazy_list

__program__ = "catastrophy"
__version__ = "0.1.0"
__authors__ = ", ".join(["Darcy Jones", "James Hane"])
__date__ = "30 March 2017"
__email__ = "darcy.a.jones@postgrad.curtin.edu.au"
__short_blurb__ = (
    "Script to predict lifestyle of filamentous plant pathogens using "
    "carbohydrate-active enzymes (CAZymes)."
)

__license__ = (
    '{__program__}-{__version__}\n'
    '{__short_blurb__}\n\n'
    'Copyright (C) {__date__},  {__authors__}'
    '\n\n'
    'This program is free software: you can redistribute it and/or modify '
    'it under the terms of the GNU General Public License as published by '
    'the Free Software Foundation, either version 3 of the License, or '
    '(at your option) any later version.'
    '\n\n'
    'This program is distributed in the hope that it will be useful, '
    'but WITHOUT ANY WARRANTY; without even the implied warranty of '
    'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the '
    'GNU General Public License for more details.'
    '\n\n'
    'You should have received a copy of the GNU General Public License '
    'along with this program. If not, see <http://www.gnu.org/licenses/>.'
).format(**locals())

EXIT_VALID = 0
EXIT_KEYBOARD = 1
EXIT_UNKNOWN = 2
EXIT_CLI = 64
EXIT_INPUT_FORMAT = 65
EXIT_INPUT_NOT_FOUND = 66
EXIT_SYSERR = 71
EXIT_CANT_OUTPUT = 73

# EXIT_IOERR = 74


class MyArgumentParser(argparse.ArgumentParser):

    def error(self, message):
        """ Override default to have more informative exit codes. """
        self.print_usage(sys.stderr)
        raise MyArgumentError("{}: error: {}".format(self.prog, message))


class MyArgumentError(Exception):

    def __init__(self, message):
        self.message = message
        self.errno = EXIT_CLI

        # This is a bit hacky, but I can't figure out another way to do it.
        if "No such file or directory" in message:
            if "infile" in message:
                self.errno = EXIT_INPUT_NOT_FOUND
            elif "outfile" in message:
                self.errno = EXIT_CANT_OUTPUT
        return


def cli(prog, args):
    parser = MyArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "Examples:\n\n"
            "```bash\n"
            "$ %(prog)s -i dbcan_results.csv -o prediction.csv\n"
            "$ %(prog)s -i dbcan_results_1.csv dbcan_results_2.csv "
            "-l result1 result2 -o prediction.csv\n"
            "```\n"
        ),
        epilog=(
            "Exit codes:\n\n"
            f"{EXIT_VALID} - Everything's fine\n"
            f"{EXIT_KEYBOARD} - Keyboard interrupt\n"
            f"{EXIT_CLI} - Invalid command line usage\n"
            f"{EXIT_INPUT_FORMAT} - Input format error\n"
            f"{EXIT_INPUT_NOT_FOUND} - Cannot open the input\n"
            f"{EXIT_SYSERR} - System error\n"
            f"{EXIT_CANT_OUTPUT} - Can't create output file\n"
            f"{EXIT_UNKNOWN} - Unhandled exception, please file a bug!\n"
            "\n"
            "Codes loosely based on <https://stackoverflow.com/questions/1101957/are-there-any-standard-exit-status-codes-in-linux>"
        )
    )

    parser.add_argument(
        "-i", "--infile",
        dest="inhandles",
        default=[sys.stdin],
        nargs="+",
        type=argparse.FileType('r'),
        help=(
            "Path to the input file output by HMMER or formatted by dbCAN "
            "`hmmscan-parser.sh`. You can specify more than one file by "
            "separating them with a space. Default is STDIN."
        )
    )

    parser.add_argument(
        "-f", "--format",
        dest="file_format",
        type=FileType.from_string,
        default=FileType.hmmer_text,
        choices=list(FileType),
        help=(
            "The format that the input is provided in. If multiple files are "
            "specified, all input must be in the same format. HMMER raw "
            "(hmmer_text, default) and domain table (hmmer_domtab) formatted "
            "files are accepted. Files processed by the dbCAN formatter "
            "`hmmscan-parser.sh` are also accepted using the `dbcan` option."
        )
    )

    parser.add_argument(
        "-l", "--label",
        dest="labels",
        default=None,
        nargs="+",
        help=(
            "Label to give the prediction for the input file. Specify more "
            "than one label by separating them with a space. The number of "
            "labels should be the same as the number of input files. "
            "By default, the filenames are used as labels."
        )
    )

    parser.add_argument(
        "-o", "--outfile",
        dest="outhandle",
        default=sys.stdout,
        type=argparse.FileType('w'),
        help=(
            "File path to write tab delimited output to. Default is STDOUT."
        )
    )

    parser.add_argument(
        "-m", "--model",
        dest="model_version",
        default=Version.latest(),
        type=Version.from_string,
        choices=list(Version),
        help=(
            "The version of the model to use. If you're using old dbCAN "
            "predictions you may have to specify this. The version numbers "
            "are just the versions of dbCAN used to train the models so just "
            "select the dbCAN version that you used. The latest version is "
            "used by default."
        )
    )

    parser.add_argument(
        "-n", "--nomenclature",
        dest="nomenclature",
        default=Nomenclature.default(),
        type=Nomenclature.from_string,
        choices=list(Nomenclature),
        help=(
            "The nomenclature type to use. "
            "Nomenclature1 is the classical symbiont, saprotroph, "
            "(hemi)biotroph, necrotroph system. Nomenclature2 separates "
            "wilts from necrotrophs, and considers symbionts as a class "
            "of biotroph. Nomenclature3 is the system proposed in the "
            "paper (default)."
        )
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the version of %(prog)s and exit"
    )

    return parser.parse_args()


def runner(inhandles, outhandle, labels, file_format,
           model_version, nomenclature):
    """ Runs the pipeline. """

    parsed = [
        parsers.parse(h, format=file_format, version=model_version)
        for h
        in inhandles
    ]
    required_cols = cazy_list(model_version)
    counts = cazy_counts_multi(parsed, labels, required_cols)
    preds = predict(
        counts,
        version=model_version,
        nomenclature=nomenclature
    )
    preds.write_tsv(outhandle)
    return


def main():
    """ The cli interface to CATAStrophy. """

    try:
        args = cli(prog=sys.argv[0], args=sys.argv[1:])
    except MyArgumentError as e:
        print(e.message, file=sys.stderr)
        sys.exit(e.errno)

    infile_names = [f.name for f in args.inhandles]

    if args.labels is None:
        labels = infile_names
    elif len(args.labels) != len(args.inhandles):
        msg = (
            "argument labels and inhandles: \n"
            "When specified, the number of labels must be the same as the "
            "number of input files. Exiting.\n"
        )

        print(msg, file=sys.stderr)
        sys.exit(EXIT_CLI)
    else:
        labels = args.labels

    try:
        runner(
            args.inhandles,
            args.outhandle,
            labels,
            args.file_format,
            args.model_version,
            args.nomenclature
        )

    except ParseError as e:
        if e.line is not None:
            header = "Failed to parse file <{}> at line {}.\n".format(
                e.filename, e.line)
        else:
            header = "Failed to parse file <{}>.\n".format(e.filename)

        print("{}\n{}".format(header, e.message), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except OSError as e:
        msg = (
            "Encountered a system error.\n"
            "We can't control these, and they're usually related to your OS.\n"
            "Try running again.\n"
        )
        print(msg, file=sys.stderr)
        print(e.strerror, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except MemoryError:
        msg = (
            "Ran out of memory!\n"
            "Catastrophy shouldn't use much RAM, so check other "
            "processes and try running again."
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except KeyboardInterrupt:
        print("Received keyboard interrupt. Exiting.", file=sys.stderr)
        sys.exit(EXIT_KEYBOARD)

    except Exception as e:
        msg = (
            "I'm so sorry, but we've encountered an unexpected error.\n"
            "This shouldn't happen, so please file a bug report with the "
            "authors.\nWe will be extremely grateful!\n\n"
            "You can email us at {}.\n"
            "Alternatively, you can file the issue directly on the repo "
            "<https://bitbucket.org/ccdm-curtin/catastrophy/issues>\n\n"
            "Please attach a copy of the following message:"
        ).format(__email__)
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)

    return


if __name__ == '__main__':
    main()



import pytest

from catas.data import Version
from catas.data import test_dbcan
from catas.parsers import split_hmm
from catas.parsers import DBCAN


@pytest.mark.parametrize("hmm,exp", [
    ("GH80.hmm", "GH80"),
    ("AC1.hmm", "AC1")
])
def test_split_hmm(hmm, exp):
    result = split_hmm(hmm)
    assert result == exp
    return


@pytest.mark.parametrize("version,idx,col,exp_val", [
    (Version.v4, 0, "hmm", "CBM1"),
    (Version.v4, 1, "hmm_len", 303),
    (Version.v4, 2, "seqid", "tr|B6DQK8|B6DQK8_PHAND"),
    (Version.v4, 3, "query_len", 637),
    (Version.v4, 6, "hmm", "GH10"),
    (Version.v4, 11, "seqid", "tr|Q9Y7H9|Q9Y7H9_PHAND"),
    (Version.v5, 0, "hmm", "CBM1"),
    (Version.v5, 1, "hmm_len", 303),
    (Version.v5, 2, "seqid", "tr|B6DQK8|B6DQK8_PHAND"),
    (Version.v5, 3, "query_len", 637),
    (Version.v5, 6, "hmm", "GH10"),
    (Version.v5, 11, "seqid", "tr|Q9Y7H9|Q9Y7H9_PHAND"),
    (Version.v6, 0, "hmm", "GH10"),
    (Version.v6, 1, "hmm_len", 29),
    (Version.v6, 2, "seqid", "tr|B6DQK8|B6DQK8_PHAND"),
    (Version.v6, 3, "query_len", 637),
    (Version.v6, 6, "hmm", "GH10"),
    (Version.v6, 11, "seqid", "tr|Q9Y7H9|Q9Y7H9_PHAND"),
])
def test_parse_dbcan_output(version, idx, col, exp_val):

    with open(test_dbcan(version=version), "r") as handle:
        sample = list(DBCAN.from_file(handle))

    assert getattr(sample[idx], col) == exp_val
    return



"""
Functions to count the number of specific classes of CAZymes.
"""

from collections import defaultdict
import numpy as np
from catas.matrix import Matrix


def cazy_counts_multi(handles, labels, required_cols):
    """ Computes counts for many files. """
    counts = [cazy_counts(m, required_cols) for m in handles]
    return Matrix(
        rows=labels,
        columns=required_cols,
        arr=np.row_stack(counts)
    )


def cazy_counts(matches, required_cols):
    """ Takes a file handle and counts unique occurrences of HMMs.

    Keyword arguments:
    matches -- A list or generator of DBCAN or HMMER records (required).
    required_cols -- A list of HMMs that must have counts in the output.

    Output:
    np.array -- An array of the counts in the same order as the required_cols.
    """

    # Loop through all lines and add seqids a set for each HMM that they match.
    these_hmms = defaultdict(set)
    for match in matches:
        these_hmms[match.hmm].add(match.seqid)

    row = np.zeros(len(required_cols), dtype=np.int)
    # Loop through columns and add the required hmm counts to output row.
    for i, col in enumerate(required_cols):
        row[i] = len(these_hmms[col])

    return row



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='catastrophy',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.1',

    description='A fungal trophy classifier based on CAZymes',
    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://bitbucket.org/ccdm-curtin/catastrophy',

    # Author details
    author='Darcy Jones',
    author_email='darcy.a.jones@postgrad.curtin.edu.au',

    # Choose your license
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='fungi machine-learning bioinformatics',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html

    # Sorry about the hard dependency for scikit-learn.
    # The way that the PCA model is stored is not necessarily stable across
    # versions of scikit-learn, so I have to keep it fixed.
    install_requires=[
        'numpy>=1.15.0',
        'biopython>=1.70',
        ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest', 'scipy', 'scikit-learn',
                'jupyter', 'matplotlib', 'seaborn'],
        'test': ['coverage', "pytest"],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'catas': ['data/*.json', 'data/*.csv', 'data/*.npz'],
    },
    include_package_data=True,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'catastrophy=catas:main',
        ],
    },
)
